import { ShowCases, HideCases } from "./popupElement.types";
import WUPBaseElement from "../baseElement";
import WUPPopupArrowElement from "./popupArrowElement";
declare const tagName = "wup-popup";
declare global {
    interface HTMLElementTagNameMap {
        [tagName]: WUPPopupElement;
    }
    namespace JSX {
        interface IntrinsicElements {
            /**  Popup element
             *  @see {@link WUPPopupElement} */
            [tagName]: WUP.Base.JSXProps<WUPPopupElement> & WUP.Popup.Attributes;
        }
    }
}
/** Popup element
 * @example
 * JS/TS
 * ```js
 * WUPPopupElement.$defaults.arrowEnable = true;
 *
 * const el = document.createElement('wup-popup');
 * el.$options.showCase = ShowCases.onClick | ShowCases.onFocus;
 * el.$options.target = document.querySelector('button');
 * // if placement impossible according to rules rest of possible rules will be applied
 * el.$options.placement = [
 *  WUPPopupElement.$placements.$top.$middle,
 *  WUPPopupElement.$placements.$bottom.$middle,
 *  WUPPopupElement.$placements.$bottom.$middle.$adjust, // adjust means 'ignore align to fit layout`
 *  WUPPopupElement.$placements.$bottom.$middle.$adjust.$resizeHeight, // resize means 'allow to resize to fit layout'
 * ];
 * document.body.append(el);
 * // or
 * const btn = document.querySelector('button');
 * // this is the most recomended way because $attach appends popup only by onShow and removes byHide
 * const detach = WUPPopupElement.$attach(
                    { target: btn, text: "Some text content here", showCase: ShowCases.onFocus | ShowCases.onClick },
                    (popup) => { popup.className = "popup-class-here"; }
                  )'
 *```
 * HTML
 * ```html
 * <button id="btn1">Target</button>
 * <!-- You can skip pointing attribute 'target' if popup appended after target -->
 * <wup-popup target="#btn1" placement="top-start">Some content here</wup-popup>
 * ```
 * @tutorial Troubleshooting:
 * * You can set minWidth, minHeight to prevent squizing of popup or don't use rule '.$adjust'
 * * Don't override styles: display, transform (possible to override only for animation)
 * * Don't use inline styles: maxWidth, maxHeight, minWidth, minHeight
 * * If target removed (when popup $isShown) and appended again you need to update $options.target (because $options.target cleared)
 * * Popup has overflow 'auto'; If you change to 'visible' it will apply maxWidth/maxHeight to first children (because popup must be restricted by maxSize to avoid layout issues)
 * * During the closing attr 'hide' is appended only if css-animation-duration is detected
 * * Popup can't be more than 100vw & 100vh (impossible to disable the rule)
 */
export default class WUPPopupElement<TOptions extends WUP.Popup.Options = WUP.Popup.Options, Events extends WUP.Popup.EventMap = WUP.Popup.EventMap> extends WUPBaseElement<TOptions, Events> {
    #private;
    static get observedOptions(): Array<keyof WUP.Popup.Options>;
    static get observedAttributes(): Array<LowerKeys<WUP.Popup.Options>>;
    static $placements: {
        $top: WUP.Popup.Place.EdgeFunc;
        $bottom: WUP.Popup.Place.EdgeFunc;
        $left: WUP.Popup.Place.EdgeFunc;
        $right: WUP.Popup.Place.EdgeFunc;
    };
    /** Returns placement */
    static $placementAttrs: (attr: WUP.Popup.Attributes["placement"]) => Array<WUP.Popup.Place.PlaceFunc> | undefined;
    static get $styleRoot(): string;
    static get $style(): string;
    /** Default options. Change it to configure default behavior */
    static $defaults: WUP.Popup.Defaults;
    /** Listen for target according to showCase and create/remove popup when it's required (by show/hide).
     *  This helps to avoid tons of hidden popups on HTML;
     *  Firing detach doesn't required if target removed by target.remove() or target.parent.removeChild(target);
     *  If target is removed via changing innerHTML you should fire detach() to avoid memoryLeak
     *  @returns detach-function (hide,remove popup and remove eventListeners)
     *  @example
     *  const detach = WUPPopupElement.$attach(
     *     {
     *       target: document.querySelector("button") as HTMLElement,
     *       text: "Some text here",
     *       showCase: ShowCases.onClick,
     *     },
     *     // (el) => el.class = "popup-attached"
     *   );
     * @tutorial Troubleshooting:
     * * $attach doesn't work with showCase.always it doesn't make sense
     * * every new attach on the same target > re-init previous (1 attach per target is possible)
     * * Firing detach() doesn't required if target removed by `target.remove()` or `target.parent.removeChild(target)`;
     * * If popup is hidden and target is removed via `target.parent.innerHTML="another content"` you should fire detach() to avoid memoryLeak
     */
    static $attach<T extends WUPPopupElement>(options: WUP.Popup.AttachOptions, 
    /** Fires when popup is added to document */
    callback?: (el: T) => void): () => void;
    /** Fires before show is happened;
     * @tutorial rules
     * * can be prevented via e.preventDefault()
     * * use event.detail.showCase to filter by showCase */
    $onWillShow?: (e: CustomEvent) => void;
    /** Fires after popup is shown (after animation finishes) */
    $onShow?: (e: Event) => void;
    /** Fires before hide is happened;
     * @tutorial rules
     * * can be prevented via e.preventDefault()
     * * use event.detail.hideCase to filter by hideCase */
    $onWillHide?: (e: CustomEvent) => void;
    /** Fires after popup is hidden (after animation finishes) */
    $onHide?: (e: Event) => void;
    /** Hide popup
     * @returns Promise resolved true if successful by animation-end */
    $hide(): Promise<boolean>;
    /** Show popup
     * @returns Promise resolved by animation-end
     * @throws (rejects) error if popup not appended on layout */
    $show(): Promise<boolean>;
    /** Force to update position. Call this if related styles are changed & need to recalc position */
    $refresh(): void;
    /** @deprecated use `$isShown` */
    get $isOpen(): boolean;
    /** Returns if popup is opened (before show-animation is started)
     * @tutorial Troubleshooting
     * * stack: $show() > `$isShown:true` > showing > shown
     * * stack: $hide() > hiding > hidden > `$isShown:false`
     * * to listen to animation-end use events `$show` & `$hide` OR methods `$show().then(...)` & `$hide().then(... )` */
    get $isShown(): boolean;
    /** Returns if popup is closed (after hide-animation is ended)
     * @tutorial Troubleshooting
     * * stack: $show() > `$isHidden:false` >  showing > shown
     * * stack: $hide() > hiding > hidden > `$isHidden:true`
     * * to listen to animation-end use events `$show` & `$hide` OR methods `$show().then(...)` & `$hide().then(... )` */
    get $isHidden(): boolean;
    /** Returns if popup is hiding (only if animation enabled) */
    get $isHiding(): boolean;
    /** Returns if popup is showing (only if animation enabled) */
    get $isShowing(): boolean;
    /** Returns arrowElement if $options.arrowEnable=true and after popup $isShown */
    get $refArrow(): WUPPopupArrowElement | null;
    protected gotReady(): void;
    /** Called after gotReady() and $show() (to reinit according to options) */
    protected init(): void;
    protected gotChanges(propsChanged: Array<string> | null): void;
    protected connectedCallback(): void;
    /** Defines target on show; @returns Element | Error */
    defineTarget(): HTMLElement | SVGElement;
    protected setMaxHeight(px: number | null): void;
    protected setMaxWidth(px: number | null): void;
    /** Collect/calc all required values into #state (when menu shows) */
    protected buildState(): void;
    /** Required to stop previous animations/timeouts (for case when option animation is changed) */
    _stopAnimation?: () => void;
    protected goAnimate(animTime: number, isHidden: boolean): Promise<boolean>;
    /** Shows popup if target defined; returns true if successful */
    goShow(showCase: ShowCases, ev: MouseEvent | FocusEvent | null): boolean | Promise<boolean>;
    /** Hide popup. @hideCase as reason of hide(). Calling 2nd time at once will stop previous hide-animation */
    goHide(hideCase: HideCases, ev: MouseEvent | FocusEvent | KeyboardEvent | null): boolean | Promise<boolean>;
    /** Returns `target.getBoundingClientRect()` Use function to change placement logic based on target-rect
     * @WARN it's called with screen-frequency (per frame) */
    getTargetRect(target: Element): DOMRect;
    /** Update position of popup. Call this method in cases when you changed options */
    protected updatePosition(): DOMRect | undefined;
    protected gotRemoved(): void;
    /** Call when need to reinit */
    protected disposeListener(): void;
    protected dispose(): void;
}
export {};
