import onFocusGot from "../helpers/onFocusGot";
import onFocusLost from "../helpers/onFocusLost";
import { focusFirst, onEvent } from "../indexHelpers";
/** Listen for target according to showCase. If target removed call stopListen + don't forget to remove popup yourself;
 * @tutorial Troubleshooting/rules:
 * * To allow user hide/show popup before previous action is ended provide for onShow and onHide-
 *   callbacks results as fast as it's possible without waiting for animation */
export default class PopupListener {
    options;
    onShow;
    onHide;
    static $defaults = {
        showCase: 4 /* ShowCases.onClick */,
        hoverShowTimeout: 200,
        hoverHideTimeout: 500,
    };
    constructor(options, onShow, onHide) {
        this.options = options;
        this.onShow = onShow;
        this.onHide = onHide;
        this.options = { ...PopupListener.$defaults, ...options };
        const t = this.options.target;
        if (!(t instanceof HTMLElement)) {
            throw new Error("WUP-Popup. Target is required");
        }
        if (!t.isConnected) {
            throw new Error("WUP-Popup. Target must be appended to layout");
        }
        this.listen();
    }
    openedEl = null;
    isShowing; // required to prevent infinite-calling from parent
    isHiding; // required to prevent infinite-calling from parent
    /** Called on show */
    async show(showCase, e) {
        if (this.openedEl || this.isShowing || !this.options.target.isConnected) {
            return;
        }
        this.isShowing = true;
        this.#lastActive = document.activeElement;
        // eslint-disable-next-line no-async-promise-executor
        this.isShowing = new Promise(async (res, rej) => {
            try {
                this.openedEl = await this.onShow(showCase, e);
                res();
            }
            catch (err) {
                rej(err);
            }
        }).finally(() => (this.isShowing = false));
        await this.isShowing;
        // timeout required to avoid immediate hide by bubbling events to root
        if (this.openedEl) {
            setTimeout(() => this.openedEl && this.listenOnShow());
        }
    }
    /** Called on hide */
    async hide(hideCase, e) {
        this.isShowing && (await this.isShowing);
        if (!this.openedEl || this.isHiding) {
            return;
        }
        this.disposeOnHide();
        const was = this.openedEl; // required when user clicks again during the hiding > we need to show in this case
        this.openedEl = null;
        try {
            this.isHiding = true;
            // prevent focusing popupContent during the hiding
            const r = onEvent(was, "focusin", ({ relatedTarget }) => {
                this.focusBack(relatedTarget); // WARN: it returns focus to target but better move focus forward - but it's impossible
            });
            const isOk = await this.onHide.call(was, hideCase, e || null);
            r();
            if (isOk) {
                this.#openedByHover = false;
                // case1: popupClick > focus lastActive or target
                // case2: hide & focus in popup > focus lastActive or target
                const isMeFocused = this.isMe(document.activeElement, was);
                const needFocusBack = isMeFocused || hideCase === 4 /* HideCases.onPopupClick */;
                needFocusBack && this.focusBack(null);
                this.#lastActive = undefined;
            }
            if (!isOk && !this.openedEl) {
                this.openedEl = was; // rollback if onHide was prevented and onShow wasn't called again during the hiding
                this.listenOnShow();
            }
        }
        catch (error) {
            Promise.reject(error); // handle error from onHide
        }
        this.isHiding = false;
    }
    /** Returns focus back during the hiding */
    focusBack(relatedTarget) {
        this.#preventFocusEvent = true;
        const t = this.options.target;
        const el = relatedTarget ||
            (t === this.#lastActive || this.includes.call(t, this.#lastActive) ? this.#lastActive : t); // focus target on item inside target if popup was focused;
        el !== document.activeElement && focusFirst(el);
        this.#preventFocusEvent = false;
    }
    /** Returns whether el is part of another  */
    includes(el) {
        return el instanceof Node && this.contains(el);
    }
    /** Returns whether element if part of openedElement */
    isMe(el, openedEl = this.openedEl) {
        return openedEl === el || this.includes.call(openedEl, el);
    }
    #preventFocusEvent;
    #preventClickAfterFocus;
    #wasOutsideClick; // fix when labelOnClick > inputOnClick
    /** Single event handler */
    handleEventsDocument(e) {
        if (e.defaultPrevented) {
            return;
        }
        // eslint-disable-next-line default-case
        switch (e.type) {
            case "focusin":
                if (!this.isMe(e.target)) {
                    this.#lastActive = e.target; // case when focus changes inside target
                }
                break;
            case "click":
                {
                    this.#preventClickAfterFocus = false; // mostly it doesn't make sense but maybe it's possible
                    // filter click from target because we have target event for this
                    const isTarget = this.includes.call(this.options.target, e.target);
                    if (isTarget) {
                        return;
                    }
                    if (this.isMe(e.target)) {
                        this.hide(4 /* HideCases.onPopupClick */, e);
                    }
                    else {
                        this.hide(3 /* HideCases.onOutsideClick */, e);
                        this.#wasOutsideClick = true;
                        setTimeout(() => (this.#wasOutsideClick = false), 50);
                    }
                }
                break;
            case "keydown":
                if (e.key === "Escape") {
                    this.hide(8 /* HideCases.onPressEsc */, e);
                }
                break;
        }
    }
    #openedByHover;
    #debounceClick;
    #hoverTimeout;
    /** Single event handler */
    handleEvents(ev) {
        if (ev.defaultPrevented) {
            return;
        }
        // eslint-disable-next-line default-case
        switch (ev.type) {
            case "click":
                {
                    const e = ev;
                    if (!e.pageX) {
                        // pageX is null or 0 if it was called programmatically
                        this.#preventClickAfterFocus = false; // test-case: focus without click > show....click programatically on target > it should hide
                    }
                    const skip = this.#preventClickAfterFocus ||
                        this.#debounceClick ||
                        this.#wasOutsideClick ||
                        this.#openedByHover ||
                        // e.detail >= 2 || // it's double-click
                        e.button; // it's not left-click
                    if (skip) {
                        return;
                    }
                    // popupClick possible when popup inside target
                    this.openedEl
                        ? this.hide(this.isMe(e.target) ? 4 /* HideCases.onPopupClick */ : 5 /* HideCases.onTargetClick */, e)
                        : this.show(4 /* ShowCases.onClick */, e);
                    // fix when labelOnClick > inputOnClick > inputOnFocus
                    this.#debounceClick = setTimeout(() => (this.#debounceClick = undefined), 1);
                }
                break;
            case "mouseenter":
                this.#hoverTimeout && clearTimeout(this.#hoverTimeout);
                if (!this.openedEl) {
                    this.#hoverTimeout = setTimeout(() => {
                        this.#openedByHover && clearTimeout(this.#openedByHover);
                        this.#openedByHover = setTimeout(() => (this.#openedByHover = false), 300); // allow hide by click when shown by hover (300ms to prevent unexpected click & hover colision)
                        this.show(1 /* ShowCases.onHover */, ev);
                    }, this.options.hoverShowTimeout);
                }
                break;
            case "mouseleave":
                this.#hoverTimeout && clearTimeout(this.#hoverTimeout);
                if (this.openedEl) {
                    this.#hoverTimeout = setTimeout(() => {
                        this.#openedByHover && clearTimeout(this.#openedByHover);
                        this.#openedByHover = false;
                        this.hide(1 /* HideCases.onMouseLeave */, ev);
                    }, this.options.hoverHideTimeout);
                }
                break;
            case "focusin":
                if (!this.openedEl && !this.#preventFocusEvent && !this.#openedByHover) {
                    this.#preventClickAfterFocus = !!(this.options.showCase & 4 /* ShowCases.onClick */);
                    this.show(2 /* ShowCases.onFocus */, ev).then(() => {
                        if (this.#preventClickAfterFocus) {
                            // WARN: event is self-destroyed on 1st event so not required to dispose it
                            document.addEventListener("pointerdown", () => (this.#preventClickAfterFocus = false), {
                                once: true,
                                capture: true,
                                passive: true,
                            }); // pointerdown includes touchstart & mousedown
                        }
                    });
                }
                break;
            case "focusout":
                /* istanbul ignore else - case impossible but better always to check this */
                if (this.openedEl) {
                    const e = ev;
                    const next = e.relatedTarget || document.activeElement;
                    const isToMe = this.isMe(next);
                    const t = this.options.target;
                    const isToTarget = t === next || this.includes.call(t, next);
                    !isToMe && !isToTarget && this.hide(2 /* HideCases.onFocusOut */, e);
                }
                break;
        }
    }
    /** Last focused element before popup is opened */
    #lastActive;
    #defargs = { passive: true };
    #disposeFocus;
    #disposeFocusLost;
    #disposeFocusLost2;
    /** Called on init to apply eventListeners */
    listen() {
        this.handleEventsDocument = this.handleEventsDocument.bind(this);
        this.handleEvents = this.handleEvents.bind(this);
        const { target: t, showCase } = this.options;
        if (showCase & 4 /* ShowCases.onClick */) {
            t.addEventListener("click", this.handleEvents, this.#defargs);
        }
        if (showCase & 1 /* ShowCases.onHover */) {
            t.addEventListener("mouseenter", this.handleEvents, this.#defargs);
            t.addEventListener("mouseleave", this.handleEvents, this.#defargs); // use only appendEvent; with onShowEvent it doesn't work properly (because filtered by timeout)
        }
        if (showCase & 2 /* ShowCases.onFocus */) {
            this.#disposeFocus = onFocusGot(t, this.handleEvents, {
                debounceMs: this.options.focusDebounceMs,
            });
            // isAlreadyFocused
            const a = document.activeElement;
            if (a === t || t.contains(a)) {
                this.handleEvents(new FocusEvent("focusin"));
                this.#preventClickAfterFocus = false;
            }
        }
    }
    /** Called to add extra events when show */
    listenOnShow() {
        const el = this.openedEl;
        document.addEventListener("focusin", this.handleEventsDocument, this.#defargs);
        document.addEventListener("keydown", this.handleEventsDocument, this.#defargs);
        const { target: t, showCase } = this.options;
        if (showCase & 4 /* ShowCases.onClick */) {
            document.addEventListener("click", this.handleEventsDocument, this.#defargs);
        }
        if (showCase & 1 /* ShowCases.onHover */) {
            el.addEventListener("mouseenter", this.handleEvents, this.#defargs);
            el.addEventListener("mouseleave", this.handleEvents, this.#defargs);
        }
        // if (showCase & ShowCases.onFocus) {
        // WARN: use focusLost forever for accessibility purpose
        this.#disposeFocusLost = onFocusLost(t, this.handleEvents, {
            debounceMs: this.options.focusDebounceMs,
        });
        this.#disposeFocusLost2 = onFocusLost(el, this.handleEvents, {
            debounceMs: this.options.focusDebounceMs,
        });
        // }
    }
    /** Called to remove extra events when hide */
    disposeOnHide() {
        document.removeEventListener("keydown", this.handleEventsDocument, this.#defargs);
        document.removeEventListener("focusin", this.handleEventsDocument, this.#defargs);
        document.removeEventListener("click", this.handleEventsDocument, this.#defargs);
        this.openedEl?.removeEventListener("mouseenter", this.handleEvents, this.#defargs);
        this.openedEl?.removeEventListener("mouseleave", this.handleEvents, this.#defargs);
        this.#disposeFocusLost?.call(this);
        this.#disposeFocusLost2?.call(this);
    }
    /** Remove all event listeners */
    stopListen() {
        this.disposeOnHide();
        const t = this.options.target;
        t.removeEventListener("click", this.handleEvents, this.#defargs);
        t.removeEventListener("mouseenter", this.handleEvents, this.#defargs);
        t.removeEventListener("mouseleave", this.handleEvents, this.#defargs);
        this.#disposeFocus?.call(this);
        this.openedEl = null;
    }
}
